: ------------------------------------------------------------------------------
: 3D environment display demo. uses double buffered video to render, flicker
: free, a simple 3D environment consisting of a bounding box and a simple flat
: plane object
:
: The environment x, y and z axes are highlit red, green and blue respectively,
: the object is rendered in magenta and highlit when the front side is visible.
: Also rendered in yellow is the object plane normal which is used to determine
: when the front face of the object is visible. This has been clipped to lie
: wholly within the environment box
:
: Use the cursor keys to rotate the environment and move it up and down and use
: the page up and page down keys to move the environment away or closer. Use
: the escape key to exit
:
:	This version for Sim68K 3.2 or later
:
:	More 68000 and other projects can be found on my website at ..
:
:	 http://www.themotionstore.com/leeedavison/index.html
:
:	.. or ..
:
:	 http://members.lycos.co.uk/leeedavison/index.html
:
:	mail : leeedavison@lycos.co.uk
: ------------------------------------------------------------------------------

NEW

:  double buffered video mode
10 MODE 17

:  set screen centre
12 xctr& = 320
14 yctr& = 120

:  set field of view angle. the field of view angle can be calculated using ..
:
:	 = ATN(xctr& / fov&) * 2
:
:  assuming  is half the screen width. for this demo it is ATN(320 / 500) * 2
:  which is nearly 66 degrees. to set a given field of view use ..
:
:	fov& = TAN( / 2) * xctr&
:
: so for a 90 degree field of view use TAN(90 / 2) * 320 which is 320

16 fov& = 500

:  set rotation step size
18 pi20 = PI / 20

:  set the environment box +/- limits
20 xsze& = $80
22 ysze& = $80
24 zsze& = $80

:  set initial environment rotation about Y axis
26 yaw = pi20 * 3
:  set initial environment origin
28 ebox& = 0
30 eboy& = 500
32 eboz& = -80

: set the three co-ordinate points [J], [K] and [L] of the plane object, in
: widdershins order as viewed from the front face

:  [J]
34 jx& = -xsze& / 2
36 jy& = ysze& / 2
38 jz& = -zsze&
:  [K]
40 kx& = 0
42 ky& = ysze&
44 kz& = 0
:  [L]
46 lx& = xsze&
48 ly& = -ysze&
50 lz& = zsze& / 2

: ------------------------------------------------------------------------------
: main loop, draw the scene then poll the movement keys until a key is detected
: then loop and do the whole thing again

100 DO

:     draw the scene
102   GOSUB 30000

:     test [esc][crup][crdn]
104   a& = KEYS($1B002628)
:     test [PgUp][PgDn][crlt][crrt]
106   b& = KEYS($21222527) : IF (a& OR b&) = 0 THEN 104

:     [cursor up] key, if less than minimum, set minimum
108   IF a& AND $FF00 THEN eboz& = eboz& - 30 : IF eboz& < -500 THEN eboz& = -500
:     [cursor down] key, if more than maximum, set maximum
110   IF a& AND $FF THEN eboz& = eboz& + 30 : IF eboz& > 100 THEN eboz& = 100

:     [PgUp] key, move away, if less than maximum distance, set maximum distance
112   IF b& AND $FF000000 THEN eboy& = eboy& * 1.25 : IF eboy& > 3000 THEN eboy& = 3000
:     [PgDn] key, move in, if more than minimum distance, set minimum distance
114   IF b& AND $00FF0000 THEN eboy& = eboy& / 1.25 : IF eboy& < 200 THEN eboy& = 200
:     [cursor left] key, rotate deocil, if > +180 degrees, make > -180 degrees
116   IF b& AND $0000FF00 THEN yaw = yaw + pi20 : IF yaw > PI THEN yaw = yaw - TWOPI
:     [cursor right] key, rotate widdershins, if < -180 degrees, make < +180 degrees
118   IF b& AND $000000FF THEN yaw = yaw - pi20 : IF yaw < - PI THEN yaw = yaw + TWOPI

120 LOOP UNTIL a& < 0

999 END

: ------------------------------------------------------------------------------
: draw the 3D environment

30000 CLS

: calculate sin and cos for the object translation
30002 sina = SIN(yaw)
30004 cosa = -COS(yaw)

: draw a bounding box at +/- x,y,z
30006 x& = -xsze& : y& = -ysze& : z& = -zsze&
30008 GOSUB 40000 : x1& = xa& : y1& = ya&
30010 x& = xsze& : y& = -ysze& : z& = -zsze&
30012 GOSUB 40000 : x2& = xa& : y2& = ya&
30014 x& = xsze& : y& = ysze& : z& = -zsze&
30016 GOSUB 40000 : x3& = xa& : y3& = ya&
30018 x& = -xsze& : y& = ysze& : z& = -zsze&
30020 GOSUB 40000 : x4& = xa& : y4& = ya&
30022 x& = -xsze& : y& = -ysze& : z& = zsze&
30024 GOSUB 40000 : x5& = xa& : y5& = ya&
30026 x& = xsze& : y& = -ysze& : z& = zsze&
30028 GOSUB 40000 : x6& = xa& : y6& = ya&
30030 x& = xsze& : y& = ysze& : z& = zsze&
30032 GOSUB 40000 : x7& = xa& : y7& = ya&
30034 x& = -xsze& : y& = ysze& : z& = zsze&
30036 GOSUB 40000 : x8& = xa& : y8& = ya&
30038 tint& = $808080
30040 LINE x1&,y1&,x2&,y2&,$0000FF
30042 LINE x2&,y2&,x3&,y3&,tint&
30044 LINE x3&,y3&,x4&,y4&,tint&
30046 LINE x4&,y4&,x1&,y1&,$00FF00

30048 LINE x5&,y5&,x6&,y6&,tint&
30050 LINE x6&,y6&,x7&,y7&,tint&
30052 LINE x7&,y7&,x8&,y8&,tint&
30054 LINE x8&,y8&,x5&,y5&,tint&

30056 LINE x4&,y4&,x8&,y8&,tint&
30058 LINE x3&,y3&,x7&,y7&,tint&
30060 LINE x2&,y2&,x6&,y6&,tint&
30062 LINE x1&,y1&,x5&,y5&,$FF0000

: ------------------------------------------------------------------------------
: create a triangular plane that is only visible from one side. the triangle is
: defined by the points [J], [K] and [L]

30064 x& = jx& : y& = jy& : z& = jz&
30066 GOSUB 40000 : x8& = xa& : y8& = ya&
:     remember the translated world co-ordinates for point [J], these are needed
:     for the visibility check later
30068 jtx& = wx& : jty& = wy& : jtz& = wz&

30070 x& = kx& : y& = ky& : z& = kz&
30072 GOSUB 40000 : x7& = xa& : y7& = ya&

30074 x& = lx& : y& = ly& : z& = lz&
30076 GOSUB 40000 : x6& = xa& : y6& = ya&

: for the triangle [J],[K],[L], where the points are in widdershins order for
: the viewed side, the vectors [a] and [b] are given by ..
:
:	[a] = [K] - [J]
:	[b] = [L] - [J]

30078 ax& = kx& - jx&
30080 ay& = ky& - jy&
30082 az& = kz& - jz&

30084 bx& = lx& - jx&
30086 by& = ly& - jy&
30088 bz& = lz& - jz&

: and the normal is given by a * b which can be calculated using ..
:
:	a * b = (ay*bz - az*by)x, (az*bx - ax*bz)y, (ax*by - ay*bx)z

30090 nx& = (ay&*bz&) - (az&*by&)
30092 ny& = (az&*bx&) - (ax&*bz&)
30094 nz& = (ax&*by&) - (ay&*bx&)

: now clip the normal to fit in the environment box and add [J] back to draw it
: relative to [J]

30096 sc = jx& / nx&

30098 x& = nx& * sc + jx&
30100 y& = ny& * sc + jy&
30102 z& = nz& * sc + jz&

30104 GOSUB 40000 : x2& = xa& : y2& = ya&

: remember the world co-ordinate normal endpoint
30106 nx& = wx& : ny& = wy& : nz& = wz&

: draw the normal

30108 tint& = $00F0F0
30110 LINE x8&,y8&,x2&,y2&,tint&

: ------------------------------------------------------------------------------
: Back face culling is when you do not render a polygon because its normal is
: pointing away from the viewer's eye point. A way of preforming this check is
: to construct a vector from the eye point to a point on the face and compute
: the dot product of this line of sight vector and the normal vector. If this
: scalar is negative then the face is pointing away and consequently can be
: culled.
:
: The dot product of two vectors [v] and [n] is given by ..
:
:	[v.n] = [v]x*[n]x + [v]y*[n]y + [v]z*[n]z
:
: .. and if the result is -ve then the face is invisible, if the result is zero
: then the face is edge on and if the result is +ve then the face is visible.

: to do this first find [v] the viewpoint normal vector from [V]-[Jt], [V] =
: 0,0,0, and the plane normal vector from [N]-[Jt]. remember these calculations
: are done using the transformed to world x,y,z co-ordinates that were saved
: earlier

30112 vx& = 0 - jtx&
30114 vy& = 0 - jty&
30116 vz& = 0 - jtz&

30118 nx& = nx& - jtx&
30120 ny& = ny& - jty&
30122 nz& = nz& - jtz&

: now find the dot product, if the result is -ve then the face is invisible

30124 dp& = vx&*nx& + vy&*ny& + vz&*nz&

: if the plane is visible draw it bright, else draw it dull

30126 tint& = $F000F0
30128 IF dp& < 0 THEN tint& = $400040

30130 LINE x8&,y8&,x7&,y7&,tint&
30132 LINE x7&,y7&,x6&,y6&,tint&
30134 LINE x6&,y6&,x8&,y8&,tint&

: copy redrawn buffer to screen

39998 BUFFER
39999 RETURN

: ------------------------------------------------------------------------------
: convert co-ordinate sets. first the object co-ordinates x,y,z are converted to
: world co-ordinates by multiplying x,y,z by the object yaw and then adding the
: object's world co-ordinate origin

40000 wx& = ebox& + (x& * cosa) - (y& * sina)
40002 wy& = eboy& + (x& * sina) + (y& * cosa)
40004 wz& = eboz& + z&

: now the world object co-ordinates are transformed to the screen co-ordinates

40006 xa& = xctr& - fov& * wx& / wy&
40008 ya& = yctr& - fov& * wz& / wy&
40010 RETURN

: ------------------------------------------------------------------------------
