*************************************************************************************
* 3D environment display demo. uses double buffered video to render, flicker		*
* free, a simple 3D environment consisting of a bounding box and a simple flat	*
* plane object												*
*														*
* The environment x, y and z axes are highlit red, green and blue respectively,	*
* the object is rendered in magenta and highlit when the front side is visible.	*
* Also rendered in yellow is the object plane normal which is used to determine	*
* when the front face of the object is visible. This has been clipped to lie		*
* wholly within the environment box									*
*														*
* Use the cursor keys to rotate the environment and move it up and down and use	*
* the page up and page down keys to move the environment away or closer. Use		*
* the escape key to exit										*
*														*
*	This version for Sim68K 3.2 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************
*
* constants

xctr	EQU		320				* screen x centre
yctr	EQU		120				* screen y centre

* set field of view angle. the field of view angle can be calculated using ..
*
*	 = ATN(xctr& / fov&) * 2
*
* assuming  is half the screen width. for this demo it is ATN(320 / 500) * 2
* which is nearly 66 degrees. to set a given field of view use ..
*
*	fov& = TAN( / 2) * xctr&
*
* so for a 90 degree field of view use TAN(90 / 2) * 320 which is 320

fov	EQU		500				* field of view
pi20	EQU		$08				* rotation step size

xsze	EQU		$80				* environment box +/- x limits
ysze	EQU		$80				* environment box +/- y limits
zsze	EQU		$80				* environment box +/- z limits

* set the three co-ordinate points [J], [K] and [L] of the plane object, in
* widdershins order as viewed from the front face

jx	EQU		-xsze/2			* [J]x
jy	EQU		ysze/2			* [J]y
jz	EQU		-zsze				* [J]z

kx	EQU		0				* [K]x
ky	EQU		ysze				* [K]y
kz	EQU		0				* [K]z

lx	EQU		xsze				* [L]x
ly	EQU		-ysze				* [L]y
lz	EQU		zsze/2			* [L]z

*************************************************************************************
*

	ORG	$1000

start
	BSR		Initialise			* go setup game stuff

	MOVEQ		#17,d1			* enable double buffering
	MOVEQ		#92,d0			* set draw mode
	TRAP		#15

	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

* main loop. everything is done in this loop, the overall speed of the game is set by
* waiting for a fixed time to pass at the end of the loop

loop
	MOVEQ		#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push time on the stack

	BSR.s		do_keys			* test and handle keypresses
	BSR		draw_3D			* go draw the scene

* now wait for a while. this limits the loop to ten redraws a second without which
* this would run way to fast if the machine is capable

	MOVE.l	(sp)+,d7			* restore time in 1/100 ths from stack
wait_100ms
	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract previous time from current time
	CMP.b		#$09,d1			* compare with 9/100ths
	BMI.s		wait_100ms			* loop if time not up yet

* if you want to see how fast it can go comment out the BMI.s above

	BRA		loop				* loop forever

*************************************************************************************
*
* test and handle keypresses

* user hit [esc]

exit
	MOVEQ		#9,d0				* exit user code
	TRAP		#15				* do I/O function

do_keys
	MOVE.l	#$1B002628,d1		* [esc][][crup][crdn] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	TST.l		d1				* test [esc] result
	BMI.s		exit				* exit if user hit [esc]

	TST.w		d1				* test [crup] result
	BPL.s		test_crdn			* if not [crup] go test [crdn]

	TST.b		d1				* test [crdn] result
	BMI.s		keys_2			* if user hit [crup] and [crdn] do nothing

	MOVEQ		#-30,d2			* set up step
	ADD.w		eboz(a3),d2			* add environment origin z
	CMP.w		#-500,d2			* compare with minimum
	BGE.s		save_eboz			* go save it if >= -500

	MOVE.w	#-500,d2			* set minimum
	BRA.s		save_eboz			* go save it

test_crdn
	TST.b		d1				* test [crdn] result
	BPL.s		keys_2			* if not [crdn] go do other keys

	MOVEQ		#30,d2			* set up step
	ADD.w		eboz(a3),d2			* add environment origin z
	CMP.w		#100,d2			* compare with maximum
	BLE.s		save_eboz			* go save it if <= 100

	MOVE.w	#100,d2			* set maximum
save_eboz
	MOVE.w	d2,eboz(a3)			* save environment origin z

keys_2
	MOVE.l	#$21222527,d1		* [PgUp][PgDn][crlt][crrt] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	TST.w		d1				* test [crlt] result
	BPL.s		test_crrt			* skip left if not pressed

	TST.b		d1				* test [crrt] result
	BMI.s		test_pgup			* if user hit [crlt] and [crrt] do nothing

	MOVEQ		#-pi20,d2			* set up step
	BRA.s		save_yaw			* go save it

test_crrt
	TST.b		d1				* test [crrt] result
	BPL.s		test_pgup			* skip right if not pressed

	MOVEQ		#pi20,d2			* set up step
save_yaw
	ADD.w		d2,yaw(a3)			* add to environment rotation about y axis

test_pgup
	SWAP		d1				* [PgUp][PgDn] into low word
	TST.w		d1				* test [PgUp] result
	BPL.s		test_pgdn			* skip up if not pressed

	TST.b		d1				* test [PgDn] result
	BMI.s		exit_keys			* if user hit [PgUp] and [PgDn] do nothing

	MOVE.w	eboy(a3),d2			* get environment origin y
	ASR.w		#2,d2				* / 4
	ADD.w		eboy(a3),d2			* add environment origin y
	CMP.w		#3000,d2			* compare with maximum
	BLE.s		save_eboy			* go save it if <= 3000

	MOVE.w	#3000,d2			* else set maximum
	BRA.s		save_eboy			* go save it

test_pgdn
	TST.b		d1				* test [PgDn] result
	BPL.s		exit_keys			* skip down if not pressed

	MOVE.w	eboy(a3),d2			* get environment origin y
	ASR.w		#1,d2				* / 2
	ADD.w		eboy(a3),d2			* add environment origin y
	ASR.w		#1,d2				* / 2, d = 3/4 * eboy(a3)
	CMP.w		#200,d2			* compare with minimum
	BGE.s		save_eboy			* go save it if >= 200

	MOVE.w	#200,d2			* set maximum
save_eboy
	MOVE.w	d2,eboy(a3)			* save environment origin z
exit_keys
	RTS


*************************************************************************************
*
* draw the 3D environment

draw_3D
	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

	BSR		sincos			* get scaled sin(yaw) in d2, cos(yaw) in d3

* draw a bounding box at +/- x,y,z

	MOVE.w	#-xsze,d4			* set point x
	MOVE.w	#-ysze,d5			* set point y
	MOVE.w	#-zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x1(a3)			* save screen x co-ordinate
	MOVE.w	d1,y1(a3)			* save screen y co-ordinate

	MOVE.w	#xsze,d4			* set point x
	MOVE.w	#-ysze,d5			* set point y
	MOVE.w	#-zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x2(a3)			* save screen x co-ordinate
	MOVE.w	d1,y2(a3)			* save screen y co-ordinate

	MOVE.w	#xsze,d4			* set point x
	MOVE.w	#ysze,d5			* set point y
	MOVE.w	#-zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x3(a3)			* save screen x co-ordinate
	MOVE.w	d1,y3(a3)			* save screen y co-ordinate

	MOVE.w	#-xsze,d4			* set point x
	MOVE.w	#ysze,d5			* set point y
	MOVE.w	#-zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x4(a3)			* save screen x co-ordinate
	MOVE.w	d1,y4(a3)			* save screen y co-ordinate

	MOVE.w	#-xsze,d4			* set point x
	MOVE.w	#-ysze,d5			* set point y
	MOVE.w	#zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x5(a3)			* save screen x co-ordinate
	MOVE.w	d1,y5(a3)			* save screen y co-ordinate

	MOVE.w	#xsze,d4			* set point x
	MOVE.w	#-ysze,d5			* set point y
	MOVE.w	#zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x6(a3)			* save screen x co-ordinate
	MOVE.w	d1,y6(a3)			* save screen y co-ordinate

	MOVE.w	#xsze,d4			* set point x
	MOVE.w	#ysze,d5			* set point y
	MOVE.w	#zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x7(a3)			* save screen x co-ordinate
	MOVE.w	d1,y7(a3)			* save screen y co-ordinate

	MOVE.w	#-xsze,d4			* set point x
	MOVE.w	#ysze,d5			* set point y
	MOVE.w	#zsze,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x8(a3)			* save screen x co-ordinate
	MOVE.w	d1,y8(a3)			* save screen y co-ordinate

	MOVEM.w	d2-d3,-(sp)			* save sin/cos

	MOVE.l	#$0000FF,d1			* set red/green/blue
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x1(a3),d1			* get line x1
	MOVE.w	y1(a3),d2			* get line y1
	MOVE.w	x2(a3),d3			* get line x2
	MOVE.w	y2(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

	MOVE.l	#$00FF00,d1			* set red/green/blue
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x4(a3),d1			* get line x1
	MOVE.w	y4(a3),d2			* get line y1
	MOVE.w	x1(a3),d3			* get line x2
	MOVE.w	y1(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

	MOVE.l	#$FF0000,d1			* set red/green/blue
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x1(a3),d1			* get line x1
	MOVE.w	y1(a3),d2			* get line y1
	MOVE.w	x5(a3),d3			* get line x2
	MOVE.w	y5(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

	MOVE.l	#$808080,d1			* set red/green/blue
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x2(a3),d1			* get line x1
	MOVE.w	y2(a3),d2			* get line y1
	MOVE.w	x3(a3),d3			* get line x2
	MOVE.w	y3(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

	MOVE.w	x3(a3),d1			* get line x1
	MOVE.w	y3(a3),d2			* get line y1
	MOVE.w	x4(a3),d3			* get line x2
	MOVE.w	y4(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x5(a3),d1			* get line x1
	MOVE.w	y5(a3),d2			* get line y1
	MOVE.w	x6(a3),d3			* get line x2
	MOVE.w	y6(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x6(a3),d1			* get line x1
	MOVE.w	y6(a3),d2			* get line y1
	MOVE.w	x7(a3),d3			* get line x2
	MOVE.w	y7(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x7(a3),d1			* get line x1
	MOVE.w	y7(a3),d2			* get line y1
	MOVE.w	x8(a3),d3			* get line x2
	MOVE.w	y8(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x8(a3),d1			* get line x1
	MOVE.w	y8(a3),d2			* get line y1
	MOVE.w	x5(a3),d3			* get line x2
	MOVE.w	y5(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x4(a3),d1			* get line x1
	MOVE.w	y4(a3),d2			* get line y1
	MOVE.w	x8(a3),d3			* get line x2
	MOVE.w	y8(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x3(a3),d1			* get line x1
	MOVE.w	y3(a3),d2			* get line y1
	MOVE.w	x7(a3),d3			* get line x2
	MOVE.w	y7(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x2(a3),d1			* get line x1
	MOVE.w	y2(a3),d2			* get line y1
	MOVE.w	x6(a3),d3			* get line x2
	MOVE.w	y6(a3),d4			* get line y2
	TRAP		#15

	MOVEM.w	(sp)+,d2-d3			* restore sin/cos

* create a triangular plane that is only visible from one side. the triangle is
* defined by the points [J], [K] and [L]

	MOVE.w	#jx,d4			* set point x
	MOVE.w	#jy,d5			* set point y
	MOVE.w	#jz,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x8(a3)			* save screen x co-ordinate
	MOVE.w	d1,y8(a3)			* save screen y co-ordinate

* remember the translated world co-ordinates for point [J], these are needed for
* the visibility check later

	MOVE.w	d4,jtx(a3)			* save world x co-ordinate
	MOVE.w	d5,jty(a3)			* save world y co-ordinate
	MOVE.w	d6,jtz(a3)			* save world z co-ordinate

	MOVE.w	#kx,d4			* set point x
	MOVE.w	#ky,d5			* set point y
	MOVE.w	#kz,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x7(a3)			* save screen x co-ordinate
	MOVE.w	d1,y7(a3)			* save screen y co-ordinate

	MOVE.w	#lx,d4			* set point x
	MOVE.w	#ly,d5			* set point y
	MOVE.w	#lz,d6			* set point z
	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x6(a3)			* save screen x co-ordinate
	MOVE.w	d1,y6(a3)			* save screen y co-ordinate

* for the triangle [J],[K],[L], where the points are in widdershins order for
* the viewed side, the vectors [a] and [b] are given by ..
*
*	[a] = [K] - [J]
*	[b] = [L] - [J]

	MOVE.w	#kx,d0			* get x
	SUB.w		#jx,d0			* subtract jx
	MOVE.w	d0,ax(a3)			* save ax
	MOVE.w	#ky,d0			* get y
	SUB.w		#jy,d0			* subtract jy
	MOVE.w	d0,ay(a3)			* save ay
	MOVE.w	#kz,d0			* get z
	SUB.w		#jz,d0			* subtract jz
	MOVE.w	d0,az(a3)			* save az

	MOVE.w	#lx,d0			* get x
	SUB.w		#jx,d0			* subtract jx
	MOVE.w	d0,bx(a3)			* save bx
	MOVE.w	#ly,d0			* get y
	SUB.w		#jy,d0			* subtract jy
	MOVE.w	d0,by(a3)			* save by
	MOVE.w	#lz,d0			* get z
	SUB.w		#jz,d0			* subtract jz
	MOVE.w	d0,bz(a3)			* save bz

* and the normal is given by a * b which can be calculated using ..
*
*	a * b = (ay*bz - az*by)x, (az*bx - ax*bz)y, (ax*by - ay*bx)z

							* nx& = (ay&*bz&) - (az&*by&)
	MOVE.w	ay(a3),d0			* get ay
	MULS.w	bz(a3),d0			* multiply by bz
	MOVE.w	az(a3),d1			* get az
	MULS.w	by(a3),d1			* multiply by by
	SUB.l		d1,d0				* subtract az*by
	LSR.l		#1,d0				* make signed word
	MOVE.w	d0,nx(a3)			* save nx

							* ny& = (az&*bx&) - (ax&*bz&)
	MOVE.w	az(a3),d0			* get az
	MULS.w	bx(a3),d0			* multiply by bx
	MOVE.w	ax(a3),d1			* get ax
	MULS.w	bz(a3),d1			* multiply by bz
	SUB.l		d1,d0				* subtract ax*bz
	LSR.l		#1,d0				* make signed word
	MOVE.w	d0,ny(a3)			* save ny

							* nz& = (ax&*by&) - (ay&*bx&)
	MOVE.w	ax(a3),d0			* get ax
	MULS.w	by(a3),d0			* multiply by by
	MOVE.w	ay(a3),d1			* get ay
	MULS.w	bx(a3),d1			* multiply by bx
	SUB.l		d1,d0				* subtract ay*bx
	LSR.l		#1,d0				* make signed word
	MOVE.w	d0,nz(a3)			* save nz

* now clip the normal to fit in the environment box and add [J] back to draw it
* relative to [J]
							* sc = jx& / nx&
	MOVEQ		#0,d0				* clear longword
	MOVE.w	#jx,d0			* get jx
	SWAP		d0				* shift to high word
	DIVS.w	nx(a3),d0			* divide by nx
	MOVE.w	d0,sc(a3)			* save scale factor

							* x& = nx& * sc + jx&
	MOVE.w	nx(a3),d4			* get nx
	MULS.w	sc(a3),d4			* * scale
	SWAP		d4				* only need high word
	ADD.w		#jx,d4			* add jx

							* y& = ny& * sc + jy&
	MOVE.w	ny(a3),d5			* get ny
	MULS.w	sc(a3),d5			* * scale
	SWAP		d5				* only need high word
	ADD.w		#jy,d5			* add jy

							* z& = nz& * sc + jz&
	MOVE.w	nz(a3),d6			* get nz
	MULS.w	sc(a3),d6			* * scale
	SWAP		d6				* only need high word
	ADD.w		#jz,d6			* add jz

	BSR		make_xy			* convert to world x,y,z and screen x,y
	MOVE.w	d0,x2(a3)			* save screen x co-ordinate
	MOVE.w	d1,y2(a3)			* save screen y co-ordinate

* remember the world co-ordinate normal endpoint

	MOVE.w	d4,nx(a3)			* save world x co-ordinate
	MOVE.w	d5,ny(a3)			* save world y co-ordinate
	MOVE.w	d6,nz(a3)			* save world z co-ordinate

* draw the normal

	MOVE.l	#$00F0F0,d1			* set red/green/blue
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x8(a3),d1			* get line x1
	MOVE.w	y8(a3),d2			* get line y1
	MOVE.w	x2(a3),d3			* get line x2
	MOVE.w	y2(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

* Back face culling is when you do not render a polygon because its normal is
* pointing away from the viewer's eye point. A way of preforming this check is
* to construct a vector from the eye point to a point on the face and compute
* the dot product of this line of sight vector and the normal vector. If this
* scalar is negative then the face is pointing away and consequently can be
* culled.
*
* The dot product of two vectors [v] and [n] is given by ..
*
*	[v.n] = [v]x*[n]x + [v]y*[n]y + [v]z*[n]z
*
* .. and if the result is -ve then the face is invisible, if the result is zero
* then the face is edge on and if the result is +ve then the face is visible.

* to do this first find [v] the viewpoint normal vector from [V]-[Jt], [V] =
* 0,0,0, and the plane normal vector from [N]-[Jt]. remember these calculations
* are done using the transformed to world x,y,z co-ordinates that were saved
* earlier

							* vx& = 0 - jtx&
	MOVEQ		#0,d0				* clear longword
	SUB.w		jtx(a3),d0			* -jtx
	MOVE.w	d0,vx(a3)			* save vx

							* vy& = 0 - jty&
	MOVEQ		#0,d0				* clear longword
	SUB.w		jty(a3),d0			* -jty
	MOVE.w	d0,vy(a3)			* save vy

							* vz& = 0 - jtz&
	MOVEQ		#0,d0				* clear longword
	SUB.w		jtz(a3),d0			* -jtz
	MOVE.w	d0,vz(a3)			* save vz

	MOVE.w	jtx(a3),d0			* get jtx&
	SUB.w		d0,nx(a3)			* nx& = nx& - jtx&
	MOVE.w	jty(a3),d0			* get jty&
	SUB.w		d0,ny(a3)			* ny& = ny& - jty&
	MOVE.w	jtz(a3),d0			* get jtz&
	SUB.w		d0,nz(a3)			* nz& = nz& - jtz&

* now find the dot product, if the result is -ve then the face is invisible

							* dp& = vx&*nx& + vy&*ny& + vz&*nz&
	MOVE.w	vx(a3),d0			* get vx
	MULS.w	nx(a3),d0			* * nx

	MOVE.w	vy(a3),d1			* get vy
	MULS.w	ny(a3),d1			* * ny

	MOVE.w	vz(a3),d2			* get vz
	MULS.w	nz(a3),d2			* * nz

	ADD.l		d1,d0				* add vy*ny
	ADD.l		d2,d0				* add vz*nz

* if the plane is visible draw it bright, else draw it dull

	MOVE.l	#$F000F0,d1			* set red/green/blue

	TST.l		d0				* test dp
	BPL.s		visible			* skip lowlight if visible

	MOVE.l	#$400040,d1			* set red/green/blue
visible
	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVE.w	x8(a3),d1			* get line x1
	MOVE.w	y8(a3),d2			* get line y1
	MOVE.w	x7(a3),d3			* get line x2
	MOVE.w	y7(a3),d4			* get line y2
	MOVEQ		#84,d0			* set draw x1,y1 to x2,y2
	TRAP		#15

	MOVE.w	x7(a3),d1			* get line x1
	MOVE.w	y7(a3),d2			* get line y1
	MOVE.w	x6(a3),d3			* get line x2
	MOVE.w	y6(a3),d4			* get line y2
	TRAP		#15

	MOVE.w	x6(a3),d1			* get line x1
	MOVE.w	y6(a3),d2			* get line y1
	MOVE.w	x8(a3),d3			* get line x2
	MOVE.w	y8(a3),d4			* get line y2
	TRAP		#15

	RTS


*************************************************************************************
*
* convert co-ordinate sets. first the object co-ordinates x,y,z, held in d4,d5,d6,
* are converted to world co-ordinates by multiplying x,y,z by the object yaw and then
* adding the object's world co-ordinate origin.
*
* returns xa,ya in d0,d1 and wx,wy,wz in d4,d5,d6

* wx& = ebox& + (x& * cosa) - (y& * sina)
* wy& = eboy& + (x& * sina) + (y& * cosa)
* wz& = eboz& + z&

make_xy
	MOVE.l	d7,-(sp)			* save register
	MOVE.w	d4,d0				* copy x co-ordinate
	MOVE.w	d5,d1				* copy y co-ordinate

							* wx = ebox + x * cos(yaw) - y * sin(yaw)
	MULS.w	d3,d4				* cos * x co-ordinate
	MULS.w	d2,d1				* sin * y co-ordinate
	SUB.l		d1,d4				* = x * cos - y * sin
	ADD.l		d4,d4				* scale 0 to 1
	SWAP		d4				* only need x high word
	ADD.w		ebox(a3),d4			* + ebox

							* wy = eboy + x * sin(yaw) + y * cos(yaw)
	MULS.w	d2,d0				* sin * x co-ordinate
	MULS.w	d3,d5				* cos * y co-ordinate
	ADD.l		d0,d5				* = x * sin + y * cos
	ADD.l		d5,d5				* scale 0 to 1
	SWAP		d5				* only need y high word
	ADD.w		eboy(a3),d5			* + eboy

							* wz = eboz + z
	ADD.w		eboz(a3),d6			* + eboz

* now the world object co-ordinates are transformed to the screen co-ordinates

	MOVE.w	d4,d0				* copy x co-ordinate
	MOVE.w	d5,d1				* copy y co-ordinate

* xa& = xctr& - fov& * wx& / wy&
* ya& = yctr& - fov& * wz& / wy&

	MULS.w	#-fov,d0			* multiply and make longword
	DIVS.w	d1,d0				* divide by distance
	ADD.w		#xctr,d0			* d0 is now screen x

	MOVE.w	d6,d7				* copy z co-ordinate

	MULS.w	#-fov,d7			* multiply and make longword
	DIVS.w	d1,d7				* divide by distance
	ADD.w		#yctr,d7			* d7 is now screen y
	MOVE.w	d7,d1				* d1 is now screen y

	MOVE.l	(sp)+,d7			* restore register
	RTS


*************************************************************************************
*
* setup stuff.

Initialise
	LEA		LAB_VARS(pc),a3		* get pointer to variables

	MOVE.w	#pi20*$1B,yaw(a3)		* set initial environment rotation about y axis

	MOVE.w	#0,ebox(a3)			* set initial environment origin x
	MOVE.w	#500,eboy(a3)		* set initial environment origin y
	MOVE.w	#-80,eboz(a3)		* set initial environment origin z

	RTS


*************************************************************************************
*
* get sin(yaw) in d2, cos(yaw) in d3.

sincos
	MOVE.w	yaw(a3),d0			* get environment orientation
	BSR.s		sin_d0			* get SIN(d0) in d0
	MOVE.w	d0,d2				* copy it

	MOVE.w	yaw(a3),d0			* get environment orientation
	BSR.s		cos_d0			* get COS(d0) in d0
	MOVE.w	d0,d3				* copy it
	RTS

*************************************************************************************
*
* get COS(d0) in d0. d0 is a nine bit value representing a full circle with the value
* increasing as you turn widdershins

cos_d0
	ADD.w		#$80,d0			* add 1/4 rotation

*************************************************************************************
*
* get SIN(d0) in d0. d0 is a nine bit value representing a full circle with the value
* increasing as you turn widdershins

sin_d0
	BTST		#8,d0				* test angle sign
	BEQ.s		cossin_d0			* just get SIN/COS and return if +ve

	BSR.s		cossin_d0			* else get SIN/COS
	NEG.w		d0				* now do twos complement
	RTS

* get d0 from SIN/COS table

cossin_d0
	TST.b		d0				* test for >= $80
	BPL.s		a_was_less			* branch if < 1/4 circle

	NEG.b		d0				* wrap $81 to $FF to $7F to $01
a_was_less
	AND.w		#$FF,d0			* ensure word high byte clear
	ADD.w		d0,d0				* * 2 bytes per word value
	LEA		sin_cos(pc),a0		* get table base address
	MOVE.w	(a0,d0.w),d0		* get SIN/COS value
	RTS

*************************************************************************************
*
* SIN/COS table, returns values between $0000 and $7FFF. the last value should be
* $8000 but that can cause an overflow in the word length calculations and it's
* easier to fudge the table a bit. no one will ever notice.

sin_cos
	dc.w	$0000,$0192,$0324,$04B6,$0648,$07D9,$096B,$0AFB
	dc.w	$0C8C,$0E1C,$0FAB,$113A,$12C8,$1455,$15E2,$176E
	dc.w	$18F9,$1A83,$1C0C,$1D93,$1F1A,$209F,$2224,$23A7
	dc.w	$2528,$26A8,$2827,$29A4,$2B1F,$2C99,$2E11,$2F87
	dc.w	$30FC,$326E,$33DF,$354E,$36BA,$3825,$398D,$3AF3
	dc.w	$3C57,$3DB8,$3F17,$4074,$41CE,$4326,$447B,$45CD
	dc.w	$471D,$486A,$49B4,$4AFB,$4C40,$4D81,$4EC0,$4FFB
	dc.w	$5134,$5269,$539B,$54CA,$55F6,$571E,$5843,$5964
	dc.w	$5A82,$5B9D,$5CB4,$5DC8,$5ED7,$5FE4,$60EC,$61F1
	dc.w	$62F2,$63EF,$64E9,$65DE,$66CF,$67BD,$68A7,$698C
	dc.w	$6A6E,$6B4B,$6C24,$6CF9,$6DCA,$6E97,$6F5F,$7023
	dc.w	$70E3,$719E,$7255,$7308,$73B6,$7460,$7505,$75A6
	dc.w	$7642,$76D9,$776C,$77FB,$7885,$790A,$798A,$7A06
	dc.w	$7A7D,$7AEF,$7B5D,$7BC6,$7C2A,$7C89,$7CE4,$7D3A
	dc.w	$7D8A,$7DD6,$7E1E,$7E60,$7E9D,$7ED6,$7F0A,$7F38
	dc.w	$7F62,$7F87,$7FA7,$7FC2,$7FD9,$7FEA,$7FF6,$7FFE
	dc.w	$7FFF

*************************************************************************************
*
* variables used

LAB_VARS						* points to variables base

	OFFSET	0				* going to use relative addressing

yaw		ds.w	1				* initial environment rotation about y axis

ebox		ds.w	1				* initial environment origin x
eboy		ds.w	1				* initial environment origin y
eboz		ds.w	1				* initial environment origin z

sc		ds.w	1				* scale factor

x1		ds.w	1				* x co-ordinate
y1		ds.w	1				* y co-ordinate
x2		ds.w	1				* x co-ordinate
y2		ds.w	1				* y co-ordinate
x3		ds.w	1				* x co-ordinate
y3		ds.w	1				* y co-ordinate
x4		ds.w	1				* x co-ordinate
y4		ds.w	1				* y co-ordinate
x5		ds.w	1				* x co-ordinate
y5		ds.w	1				* y co-ordinate
x6		ds.w	1				* x co-ordinate
y6		ds.w	1				* y co-ordinate
x7		ds.w	1				* x co-ordinate
y7		ds.w	1				* y co-ordinate
x8		ds.w	1				* x co-ordinate
y8		ds.w	1				* y co-ordinate

ax		ds.w	1				* x co-ordinate
ay		ds.w	1				* y co-ordinate
az		ds.w	1				* z co-ordinate

bx		ds.w	1				* x co-ordinate
by		ds.w	1				* y co-ordinate
bz		ds.w	1				* z co-ordinate

jtx		ds.w	1				* x co-ordinate
jty		ds.w	1				* y co-ordinate
jtz		ds.w	1				* z co-ordinate

nx		ds.w	1				* x co-ordinate
ny		ds.w	1				* y co-ordinate
nz		ds.w	1				* z co-ordinate

vx		ds.w	1				* x co-ordinate
vy		ds.w	1				* y co-ordinate
vz		ds.w	1				* z co-ordinate

	END	start

